//------------------------------------------------------------------------------
//             Copyright (c) 2002-2003, Tech-Know-Ware Ltd.
//------------------------------------------------------------------------------


// File name:   lumas-lib.c
//------------------------------------------------------------------------------
// Description: This file contains library functions for reading Lumas encoded
//				messages.
//
// Notice:      TERMS & CONDITIONS
//
//              Your use of this file implies agreement to the Terms & 
//              Conditions as presented in this notice.
//
//              Subject to the terms and conditions of this notice, 
//              TECH-KNOW-WARE LTD hereby grants YOU (the person accepting this 
//              agreement) a non-exclusive right to use this file.
//
//              THIS FILE AND THE SOFTWARE CONTAINED HEREIN IS PROVIDED 'AS IS' 
//              AND COMES WITH NO WARRENTY OF ANY KIND.  
//
//              TECH-KNOW-WARE LTD SHALL NOT BE LIABLE FOR ANY DAMAGES SUFFERED 
//              BY ANYONE OR ANYTHING DUE TO THE USE OF THIS FILE HOWEVER THEY 
//              MAY BE CAUSED.
//
//              This file, or any derivative of it, is provided on the condition
//              that YOU may not supply it to a third party outside of the 
//              organisation to which you are affiliated in a form that 
//              allows the third party to compile or link the software into 
//              programs of their own making.  Other than the previous 
//              condition, this file, or any derivative of it, may be used for 
//              any purpose, including commercial purposes. 
//
//              Users are permitted to modify the contents of this file, as 
//              long this notice and the copyright notice remains and the 
//              modified file is given a name that can not be confused with 
//              the files provided by Tech-Know-Ware Ltd.
//
//              Users are permitted to extract elements of this file into
//              other files, subject to this notice and the copyright notice 
//              also being copied into the target files.
//
//              For more information on Lumas, visit:
//
//                  http://www.tech-know-ware.com/lumas
//
//              If you find a bug in this software, please report it to:
//
//                  lumas-lib-bugs@tech-know-ware.com
//
//              Copyright (c) 2002, 2003 Tech-Know-Ware Ltd.
//
//              End of Notice.
//
// Module Tag:  lml - Lumas Library.
// Version:     V.96
// Entry Points: 
//				flml_is_message_complete
//				flml_parse_message
//				flml_set_location
//				flml_restore_location
//				flml_is_present
//				flml_read_void
//				flml_read_boolean
//				flml_read_integer
//				flml_read_float
//				flml_read_double
//				flml_read_ipv4
//				flml_read_ipv6 - Not in this release
//				flml_read_date
//				flml_read_time
//				flml_read_oid
//				flml_read_ascii
//				flml_read_ascii_dynamic
//				flml_read_const
//				flml_read_unicode
//				flml_read_unicode_dynamic
//				flml_read_bytes
//				flml_read_bytes_dynamic
//				flml_read_embedded
//				flml_read_embedded_dynamic
//				flml_read_union
//				flml_read_enumerated_union
//				flml_release
//
//				flml_encode_double
//				flml_encode_get_ascii_length
//				flml_encode_ascii
//				flml_encode_get_unicode_length
//				flml_encode_unicode
//				flml_encode_get_bytes_length
//				flml_encode_bytes


#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <float.h>


#include "lumas-lib.h"

#define LML_USE_COMPILED_TAG_INFO 0


#define KLML_EOM			(0)
#define KLML_MAX_RECURSIONS	100	/* This may need to be adjusted according to the message definition */

#define MLML_ASSERT(x)		/* Users of this code may define this */


#define MLML_IS_FIRST_TOKEN_CHAR( x ) ( \
								(x) > ' ' &&  /* No WS, no Ctrl */	\
								! strchr( ",\"'{}()[-", (x) ) &&	\
								(x) != KLML_EOM )

#define MLML_IS_FIRST_TAG_CHAR( x ) ( \
								MLML_IS_FIRST_TOKEN_CHAR( x ) &&	\
								! isdigit( (x) ) )

#define MLML_IS_TOKEN_CHAR( x ) ( \
								(x) > ' ' &&  /* No WS, no Ctrl */	\
								! strchr( "}),=", (x) ) &&			\
								(x) != KLML_EOM )

// Tests for definite value start - Values may start with other characters also (e.g. ascii)
#define MLML_IS_VALUE_START( x ) ( \
								(x) == '"' || (x) == '\'' ||		\
								(x) == '(' || (x) == '[' ||			\
								(x) == '-' || isdigit( (x) ) )


#define MLMS_AFTER_VALUE_CHAR( x ) ( isspace( (x) ) || (x) == ',' || \
									(x) == '}' || (x) == ')' || (x) == KLML_EOM )


// These are for the position data file
const char glml_is_struct[] = "s";
const char glml_is_union[] = "u";
const char glml_is_untagged[] = "";


static int p_parse( slml_parse_ctrl *ap_parse_ctrl, slml_message_element **app_new_element );
static void p_free_element_info( slml_message_element *ap_element );
tlms_bool flml_initialise( slml_parse_ctrl *ap_parse_ctrl );


//-----------------------------------------------------------------------------
//          The following functions handle reading the message input.
//     They are designed so that it is easy to alter whether the message is
//                      stored in memory or in a file.
//-----------------------------------------------------------------------------


static tlml_input_location p_get_input_location( slml_parse_ctrl *ap_parse_ctrl )
{
    return ap_parse_ctrl->p_next - ap_parse_ctrl->p_message;
}


static void p_set_input_location( slml_parse_ctrl *ap_parse_ctrl,
									tlml_input_location a_location )
{
    ap_parse_ctrl->p_next = ap_parse_ctrl->p_message + a_location;
	ap_parse_ctrl->unget_char = '\0';
}


// Function name:   p_unget_char
//-----------------------------------------------------------------------------
// Description: Ungetting a char is handled by storing the ungot char rather
//				than by adjusting the ap_parse_ctrl->p_next pointer because
//				the char being returned could be a space substituted
//				for a comment.  In this case, updating ap_parse_ctrl->p_next
//				does not work.

static void p_unget_char( slml_parse_ctrl *ap_parse_ctrl, int a_c )
{
	ap_parse_ctrl->unget_char = a_c;
}


// Function name:   p_get_char
//-----------------------------------------------------------------------------
// Description: Reads next characeter from file.  This function can be called 
//				directly if comment removal is not required.

static int p_get_char( slml_parse_ctrl *ap_parse_ctrl )
{
	if( ap_parse_ctrl->unget_char != '\0' )
		{
		int l_return_char = ap_parse_ctrl->unget_char;

		ap_parse_ctrl->unget_char = '\0';

		return l_return_char;
		}

	if( ap_parse_ctrl->p_next < ap_parse_ctrl->p_message_end )
		return *(ap_parse_ctrl->p_next++);

	return KLML_EOM;
}


static int p_get_unicode_char( slml_parse_ctrl *ap_parse_ctrl )
{
	int l_char;
	int l_c = p_get_char( ap_parse_ctrl );

	if( l_c <= 0x7f )
		l_char = l_c & 0xff;

	else if( l_c <= 0xc0 )
		{
		// 2 bytes
		l_char = (l_c & 0x1f) << 6;
		l_c = p_get_char( ap_parse_ctrl );
		l_char |= l_c & 0x3f;
		}

	else if( l_c <= 0xe0 )
		{
		// 3 bytes
		l_char = (l_c & 0x1f) << 6;
		l_c = p_get_char( ap_parse_ctrl );
		l_char |= l_c & 0x3f;
		l_char <<= 6;
		l_c = p_get_char( ap_parse_ctrl );
		l_char |= l_c & 0x3f;
		}

	else
		return KLML_EOM;		// Can not do UCS-4, so stop parsing

	if( l_c == KLML_EOM )		// Test premature end of message for all cases
		return KLML_EOM;

	return l_char;
}


static int p_skip_comment_to_nl( slml_parse_ctrl *ap_parse_ctrl )
{
    int l_c;

    do  
        {
        l_c =  p_get_char( ap_parse_ctrl );   // Read next charater

        } while( l_c != '\n' && l_c != KLML_EOM );    
                                                // Until comment end

    l_c = ' ';     // Treat a comment as single white space

    return l_c;
}


static int p_skip_comment( slml_parse_ctrl *ap_parse_ctrl )
{
    int l_c;
    int l_extra = '\0';     // Initialise to some non-valid character

    do  
        {
        l_c = l_extra;								// Store used character
        l_extra = p_get_char( ap_parse_ctrl );		// Read next charater
        if( l_extra == KLML_EOM )                   // Check not at end of file
            {
            break;
            }
        } while( l_c != '*' || l_extra != '/' );    
                                                    // Until comment end

    l_c = ' ';     // Treat a comment as single white space


    return l_c;
}


// Function name: p_read
//-----------------------------------------------------------------------------
// Description: Reads input and removes any comments before passing a 
//              character that is not part of a comment back to the calling
//              function.  
//              Call p_get_char directly if it is desired not to remove things
//              that look like comments (This may occur in strings).

static int p_read( slml_parse_ctrl *ap_parse_ctrl )
{
    int     l_c;

    if( (l_c = p_get_char( ap_parse_ctrl )) == KLML_EOM )
        return KLML_EOM;


    // Check for comments - Put in while loop to allow back-to-back comments
    while( l_c == '/' )
        {
        int     l_extra;

        l_extra = p_get_char( ap_parse_ctrl );				// Read next character from file
        if( l_extra == '/' )								// If this is a comment then
            l_c = p_skip_comment_to_nl( ap_parse_ctrl );	// -skip the comment
		else if( l_extra == '*' )
			l_c = p_skip_comment( ap_parse_ctrl );
        else
            {
            p_unget_char( ap_parse_ctrl, l_extra );

            break;
            }
        }

    return l_c;                  // Character nothing to do with comment so
                                    // -return it
}


//-----------------------------------------------------------------------------
//      Start of functions for parsing message into linked list tree 
//                         used to locate parameters.
//-----------------------------------------------------------------------------

static slml_message_element *p_new_message_element()
{ 
	slml_message_element *lp_new_element = 
			(slml_message_element *)malloc( sizeof( slml_message_element ));

    lp_new_element->p_tag = NULL;
    lp_new_element->tag_location = -1;
    lp_new_element->value_location = -1;
    lp_new_element->p_compound = NULL;
    lp_new_element->p_next = NULL;

	return lp_new_element;
}


static int p_read_token( slml_parse_ctrl *ap_parse_ctrl, int a_c,
							char *ac_token, int a_token_size )
{
	int l_i;
	int l_include_equals = 0;

	MLML_ASSERT( a_token_size > 0 );

	a_token_size--;		// Make space for termination

    for( l_i=0; 
			l_i < a_token_size && MLML_IS_TOKEN_CHAR( a_c );
			l_i++ )
        {
		ac_token[l_i] = a_c;

		a_c = p_read( ap_parse_ctrl );
        }


	ac_token[l_i] = '\0';


	// Skip anything that won't fit into array: This effectively discards tokens that
	// - are too long rather than creating an exception.

	while( MLML_IS_TOKEN_CHAR( a_c ) )
		a_c = p_read( ap_parse_ctrl );

	return a_c;		// Note: KLML_EOM is a valid return value for this function
					// - and does NOT indicate an error.
}


static void p_skip_value( slml_parse_ctrl *ap_parse_ctrl, int a_c )
{
	if( a_c == '(' )
		{
		int l_nesting = 0;

		a_c = p_read( ap_parse_ctrl );

		while( a_c != KLML_EOM )
			{
			if( a_c == '(' )
				l_nesting++;

			else if( a_c == ')' )
				{
				l_nesting--;
				if( l_nesting < 0 )
					return;
				}

			else if( a_c == '=' || isspace( a_c ) || a_c == '{' || a_c == '}' )
				{
				// Ignore and move on
				}

			else
				p_skip_value( ap_parse_ctrl, a_c );		// Skip tags as well as values
		
			a_c = p_read( ap_parse_ctrl );
			}
		}
    
	else if( a_c == '"' || a_c == '\'' )
        {
        int l_end_char = a_c;
        int l_escaped = FALSE;

        while( (a_c = (l_end_char != '"') ? 
							p_get_char( ap_parse_ctrl ) : 
							p_get_unicode_char( ap_parse_ctrl )) != KLML_EOM && 
					(a_c != l_end_char || l_escaped) )
            {
            if( !l_escaped && a_c == '\\' )
                l_escaped = TRUE;

            else if( l_escaped )
                l_escaped = FALSE;
            }
        }

	else if( a_c == '[' )
        {
        while( (a_c = p_read( ap_parse_ctrl )) != KLML_EOM && a_c != ']' )
            {
            }
		}

	else
		{
        while( (a_c = p_read( ap_parse_ctrl )) != KLML_EOM && MLML_IS_TOKEN_CHAR( a_c ) )
            {
            }

		p_unget_char( ap_parse_ctrl, a_c );
		}
}


// Func name:   p_parse_value
//------------------------------------------------------------------------------
// Notes:		Locate-Correct - When we locate a value that is also a tag, we
//				have actually encountered a nested union.  It is important that
//				we set the tag_number to 0 so that the parameter location works
//				properly for this case.  See p_locate_parameter description.

static int p_parse_value( slml_parse_ctrl *ap_parse_ctrl, 
							slml_message_element *ap_element )
{
	int l_c;

	if( ap_parse_ctrl->recursion++ > KLML_MAX_RECURSIONS )
		{
		ap_parse_ctrl->error = ELML_TOO_MANY_RECURSIONS;
		return FALSE;
		}

	l_c = p_read( ap_parse_ctrl );		// Get character after '='

	while( isspace( l_c ) )
		l_c = p_read( ap_parse_ctrl );

	ap_element->value_location = p_get_input_location( ap_parse_ctrl ) - 1;

	// Handle value appropriately
    if( l_c == KLML_EOM || l_c <= ' ' || l_c == '}' || l_c == ')' || 
							l_c == '=' || l_c ==',' )
		{
		ap_parse_ctrl->error = ELML_PARSE_FAILED;

		return FALSE;
		}

	else if( MLML_IS_VALUE_START( l_c ) )
		{
		p_skip_value( ap_parse_ctrl, l_c );
		}

	else if( l_c == '{' )
		{
		if( ! p_parse( ap_parse_ctrl, &(ap_element->p_compound) ) )
			return FALSE;
		}

	else		// Any other character (and can not be white space)
		{
		char lc_value[KLML_MAX_TAG];
		tlml_input_location l_value_location = p_get_input_location( ap_parse_ctrl ) - 1;

		l_c = p_read_token( ap_parse_ctrl, l_c, lc_value, sizeof( lc_value ) );

		// Now look for possible following '='
		while( isspace( l_c ) )
			l_c = p_read( ap_parse_ctrl );

		if( l_c == '=' )
			{
			// This is something like tag = value&tag = value
			if( ! (ap_element->p_compound = p_new_message_element()) )
				{
				ap_parse_ctrl->allocation_failed = TRUE;

				return FALSE;
				}

			if( ! (ap_element->p_compound->p_tag = strdup( lc_value ) ) )
				{
				ap_parse_ctrl->allocation_failed = TRUE;

				return FALSE;
				}

			ap_element->p_compound->tag_location = l_value_location;
			ap_element->p_compound->tag_number = 0;			// See Locate-Correct

			if( ! p_parse_value( ap_parse_ctrl, ap_element->p_compound ) )
				return FALSE;
			}

		else
			p_unget_char( ap_parse_ctrl, l_c );
		}

	ap_parse_ctrl->recursion--;

	return TRUE;
}


// Func name:   p_parse
//------------------------------------------------------------------------------
// Notes:		Tag-modify - When we assign a tag number to a parameter, we
//				assume that the next parameter will not have the same tag 
//				number and increment l_tag_number accordingly.  If we 
//				subsequently encounter a comma (indicating that the next 
//				parameter should have the same tag number as the previous 
//				parameter), then we correct l_tag_number by decrementing it.
//				Note also that we have to assign the tag number after we've
//				tested for the comma so that the corrected value is used.
//				Hence the assignment is done at the end of the loop rather
//				than with all the other lp_new_element initialisation code.

static int p_parse( slml_parse_ctrl *ap_parse_ctrl, 
					slml_message_element **app_new_seg_head )
{
    int l_c;
    slml_message_element *lp_segment_tail = NULL;
    char *lp_last_tag = "";
	tlml_input_location l_last_tag_location = -1;
	int l_tag_number = 0;
	int l_first_time = TRUE;


	if( ap_parse_ctrl->recursion++ > KLML_MAX_RECURSIONS )
		{
		ap_parse_ctrl->error = ELML_TOO_MANY_RECURSIONS;
		return FALSE;
		}

    for( ;; )
        {
        char lc_tag[KLML_MAX_TAG];
        slml_message_element *lp_new_element;

        lc_tag[0] = '\0';

        while( (l_c = p_read( ap_parse_ctrl )) != KLML_EOM && isspace( l_c ) )
            {
            }

        if( l_c == '}' )
            break;

		else if( l_c == KLML_EOM )
			{
			if( ap_parse_ctrl->recursion == 1 )
				break;		// Case where transport indicates end of message

			else
				{
				ap_parse_ctrl->error = ELML_MESSAGE_TOO_SHORT;

				return FALSE;
				}
			}

		else if( l_c < ' ' || l_c == '=' || l_first_time == TRUE && l_c ==',' )
			{
			ap_parse_ctrl->error = ELML_PARSE_FAILED;

			return FALSE;
			}

		l_first_time = FALSE;

        if( ! (lp_new_element = p_new_message_element()) )
            {
            ap_parse_ctrl->allocation_failed = TRUE;

            return FALSE;
            }

        if( lp_segment_tail )
            lp_segment_tail->p_next = lp_new_element;
        else
            *app_new_seg_head = lp_new_element;

        lp_segment_tail = lp_new_element;


		if( MLML_IS_VALUE_START( l_c ) ) 		// Value with no tag
			{
			lp_new_element->value_location = p_get_input_location( ap_parse_ctrl ) - 1;
			lp_new_element->tag_number = l_tag_number;

			p_skip_value( ap_parse_ctrl, l_c );
			}

		else if( l_c == '{' )						// Value with no tag
			{
			lp_new_element->value_location = p_get_input_location( ap_parse_ctrl ) - 1;
			if( ! p_parse( ap_parse_ctrl, &(lp_new_element->p_compound) ) )
				return FALSE;
			}

		else if( l_c == ',' )
			{
			if( ! lp_last_tag )
				lp_new_element->p_tag = NULL;

			else if( ! (lp_new_element->p_tag = strdup( lp_last_tag ) ) )
				{
				ap_parse_ctrl->allocation_failed = TRUE;

				return FALSE;
				}

			lp_new_element->tag_location = l_last_tag_location;

			if( l_tag_number > 0 ) l_tag_number--;		// See Tag-modify

			if( ! p_parse_value( ap_parse_ctrl, lp_new_element ) )
				return FALSE;
			}

		else
			{
			// Read what we think is the tag
			lp_new_element->tag_location = p_get_input_location( ap_parse_ctrl ) - 1;

			l_c = p_read_token( ap_parse_ctrl, l_c, lc_tag, sizeof( lc_tag ) );

			if( ! (lp_new_element->p_tag = strdup( lc_tag ) ) )
				return FALSE;

			// Look for =
			while( isspace( l_c ) )
				l_c = p_read( ap_parse_ctrl );

			if( l_c == '=' )
				{
				// We now KNOW that we are looking for a value
				p_parse_value( ap_parse_ctrl, lp_new_element );
				}

			else
				{
				// AMBIGUOUS - What we already have is 
				// -either value without a tag OR tag without a value
				lp_new_element->value_location = lp_new_element->tag_location;
				p_unget_char( ap_parse_ctrl, l_c );
				}
			}


		lp_new_element->tag_number = l_tag_number++;		// See Tag-modify

		lp_last_tag = lp_new_element->p_tag;
		l_last_tag_location = lp_new_element->tag_location;
        }


	ap_parse_ctrl->recursion--;

	return TRUE;
}


//-----------------------------------------------------------------------------
//          Functions for debugging and seeing what is going on
//-----------------------------------------------------------------------------


static void p_display_parse_info( FILE *ah_fout, slml_message_element *ap_element, int a_indent )
{
    while( ap_element )
        {
        int l_i;
		char *lp_tag = ap_element->p_tag;

		if( ! lp_tag )
			lp_tag = "?";

        for( l_i=0; l_i<a_indent; l_i++ )
            fprintf( ah_fout, "    " );

        fprintf( ah_fout, "Name: %s, tag offset: %d, tag number: %d, value offset: %d\n",
                lp_tag,
                ap_element->tag_location,
				ap_element->tag_number,
                ap_element->value_location );

        if( ap_element->p_compound )
            p_display_parse_info( ah_fout, ap_element->p_compound, a_indent + 1 );

        ap_element = ap_element->p_next;
        }
}


void flml_debug_display_parse_info( FILE *ah_fout, slml_parse_ctrl *ap_parse_ctrl )
{
	p_display_parse_info( ah_fout?ah_fout:stdout, 
							ap_parse_ctrl->root_location.p_element, 0 );
}

//-----------------------------------------------------------------------------
//        Internal high-level functions for querying message contents
//-----------------------------------------------------------------------------

// Format of ap_tagging is: =<no. untagged><tag name>,<instance> e.g. =3mine,2

static elml_error p_locate_parameter( slml_location *ap_location,
											const char *ap_tagging,
											int a_instance )
{
    char	lc_tag_looked_for[KLML_MAX_TAG];
    int		l_i;
	int		l_untagged_number = 0;
	int		l_instance = 0;
	slml_tag_info *lp_tag_info = ap_location->p_tag_info;


	// Sort out untagged indexing first
	if( isdigit( *ap_tagging ) )
		{
		l_untagged_number = *ap_tagging++ - '0';

		while( isdigit( *ap_tagging ) )
			{
			l_untagged_number *= 10;
			l_untagged_number += *ap_tagging++ - '0';
			}
		}

    // Extract the next segment to be looked for
    for( l_i = 0; *ap_tagging != '=' && *ap_tagging != ',' && *ap_tagging != '\0'; l_i++ )
        lc_tag_looked_for[l_i] = *ap_tagging++;

    lc_tag_looked_for[l_i] = '\0';


	l_instance = 0;
	if( *ap_tagging == ',' )
		{
		++ap_tagging;

		while( isdigit( *ap_tagging ) )
			{
			l_instance *= 10;
			l_instance += *ap_tagging++ - '0';
			}
		}

	if( *ap_tagging == '\0' && a_instance >= 0 )
		l_instance = a_instance;


#if LML_USE_COMPILED_TAG_INFO != 0
	if( lc_tag_looked_for[0] != '\0' )		// If text tag, extract number of untagged items
		l_untagged_number = (int)(lp_tag_info->p_definition);

	else if( l_untagged_number >= (int)(lp_tag_info->p_definition) )	// Check requested untagged value exists
		{
		MLML_ASSERT( 0 );

		return ELML_PARAMETER_DOES_NOT_EXIST;
		}



	// Skip the appropriate number of untagged items in the tag info 
	// -structure depending on if struct or union.  

	if( lp_tag_info->p_name == glml_is_struct )
		{
		lp_tag_info++;	// Move into individual parameter specification part of data structure

		for( l_i=0; l_i < l_untagged_number; l_i++ )	// Falling off end test done above!
			lp_tag_info++;
		}

	else 
		{
		lp_tag_info++;	// Move into individual parameter specification part of data structure

		if( lc_tag_looked_for[0] == '\0' )
			{
			if( lp_tag_info->p_name != glml_is_untagged )
				{
				MLML_ASSERT( 0 );

				return ELML_PARAMETER_DOES_NOT_EXIST;
				}
			}
		else
			{
			if( lp_tag_info->p_name == glml_is_untagged )
				lp_tag_info++;
			}
		}


	if( lc_tag_looked_for[0] != '\0' )
		{
		while( lp_tag_info->p_name != glml_is_struct && 
				lp_tag_info->p_name != glml_is_union )
			{
			if( lp_tag_info->p_name != glml_is_untagged &&					// Ensures ->p_name[1] is present
							lc_tag_looked_for[0] == *(lp_tag_info->p_name) &&
							strcmp( &lc_tag_looked_for[1], lp_tag_info->p_name + 1 ) == 0 )
				{
				break;
				}

			lp_tag_info++;
			}
		}


	if( lp_tag_info->p_name == glml_is_struct || 
			lp_tag_info->p_name == glml_is_union ||
			*ap_tagging == '=' && !lp_tag_info->p_definition )
		{
		MLML_ASSERT( 0 );

		return ELML_PARAMETER_DOES_NOT_EXIST;
		}


	if( *ap_tagging == '=' )
		ap_location->p_tag_info = lp_tag_info->p_definition;
	else
		ap_location->p_tag_info = lp_tag_info;

#endif


	// Handle the untagged part.  See Locate-Correct
	while( ap_location->p_element && 
					ap_location->p_element->tag_number < l_untagged_number )
		ap_location->p_element = ap_location->p_element->p_next;


	// If we're looking for untagged, find it now
	if( lc_tag_looked_for[0] == '\0' )
		{
		// Get the right instance of the untagged value

		while( ap_location->p_element && 
						ap_location->p_element->tag_number == l_untagged_number )
			{
			if( l_instance-- == 0 )
				{
				if( *ap_tagging == '=' )
					{
					ap_location->p_element = ap_location->p_element->p_compound;

					return( p_locate_parameter( ap_location, 
												ap_tagging + 1,
												a_instance ) );
					}
				else
					return ELML_OK;
				}

			ap_location->p_element = ap_location->p_element->p_next;
			}

		return ELML_PARAMETER_NOT_FOUND;
		}

	else	// tagged item
		{
		while( ap_location->p_element )
			{
			if( ap_location->p_element->p_tag &&	// Match tagged
							lc_tag_looked_for[0] == *(ap_location->p_element->p_tag) &&
							strcmp( &lc_tag_looked_for[1], ap_location->p_element->p_tag + 1 ) == 0 )
				{
				if( l_instance-- == 0 )
					{
					if( *ap_tagging == '=' )
						{
						ap_location->p_element = ap_location->p_element->p_compound;

						return( p_locate_parameter( ap_location, 
													ap_tagging + 1,
													a_instance ) );
						}
					else
						return ELML_OK;
					}

				}

			ap_location->p_element = ap_location->p_element->p_next;
			}
		}


    return ELML_PARAMETER_NOT_FOUND;
}


static slml_message_element *p_locate( slml_parse_ctrl *ap_parse_ctrl, 
										const char *ap_tagging,
										int a_instance )
{
	slml_location l_location = ap_parse_ctrl->base_location;

	if( *ap_tagging == '=' )
		{
		l_location = ap_parse_ctrl->root_location;
		ap_tagging++;
		}

	if( ! l_location.p_element )
		{
		// No parsed data structure present
		ap_parse_ctrl->error = ELML_NO_DATA_IN_MESSAGE;
		return NULL;
		}

	ap_parse_ctrl->error = p_locate_parameter( &l_location, 
												ap_tagging,
												a_instance );

	if( ap_parse_ctrl->error != ELML_OK )
		return NULL;

	return l_location.p_element;
}


//-----------------------------------------------------------------------------
//                  Helper functions for reading the values
//-----------------------------------------------------------------------------


static tlms_bool p_get_first_char( slml_parse_ctrl *ap_parse_ctrl,
									tlml_input_location a_location,
									int *ap_c )
{
    p_set_input_location( ap_parse_ctrl, a_location );

    while( (*ap_c = p_read( ap_parse_ctrl )) != KLML_EOM && isspace( *ap_c ) )
        {
        }


    if( *ap_c == KLML_EOM )
		{
		ap_parse_ctrl->error = ELML_MESSAGE_TOO_SHORT;
		return FALSE;
		}

	return TRUE;
}


static tlms_bool p_read_pos_integer( slml_parse_ctrl *ap_parse_ctrl, unsigned int *ap_int )
{
	unsigned int l_integer = 0;
	int l_c = p_read( ap_parse_ctrl );

	if( !isdigit( l_c ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	do
		{
		l_integer = 10 * l_integer + l_c - '0';

		l_c = p_read( ap_parse_ctrl );

		} while( isdigit( l_c ) );

	p_unget_char( ap_parse_ctrl, l_c );

	*ap_int = l_integer;

	return TRUE;
}


static tlms_bool p_is_valid_float( const char *ap_value )
{
	// Pattern [-] 1*DIGIT [ "." *DIGIT ][ ("e" / "E") ["+"/"-"] 1*DIGIT ) ] 

	if( *ap_value == '-' )			
		ap_value++;

	if( ! isdigit( *ap_value ) )
		return FALSE;

	ap_value++;

	while( isdigit( *ap_value ) )
		ap_value++;

	if( *ap_value == '.' )
		{
		ap_value++;

		while( isdigit( *ap_value ) )
			ap_value++;
		}

	if( *ap_value == 'e' || *ap_value == 'E' )
		{
		ap_value++;
		if( *ap_value == '+' || *ap_value == '-' )
			ap_value++;

		if( ! isdigit( *ap_value ) )
			return FALSE;

		while( isdigit( *ap_value ) )
			ap_value++;
		}

	if( *ap_value == '\0' )
		return TRUE;

	return FALSE;
}


static double p_float_zero()
{
	return 0.0;
}


static tlms_bool p_get_ascii_length( slml_parse_ctrl *ap_parse_ctrl,
										slml_message_element *ap_element,
										int *ap_length )
{
    int l_c;
    int l_length;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( ! (l_c == '\'' || MLML_IS_FIRST_TOKEN_CHAR( l_c ) ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	// Note direct call to p_get_char so that contents of string can not be interpreted
	// -as comments

	if( l_c != '\'' )
		{
		for( l_length = 0; MLML_IS_TOKEN_CHAR( l_c ); l_length++ )
			{
			l_c = p_get_char( ap_parse_ctrl );
			}
		}
	else
		{
		int l_escaped = FALSE;
	    l_length = 0;

		while( (l_c = p_get_char( ap_parse_ctrl )) != KLML_EOM && 
						(l_c != '\'' || l_escaped) )
			{
			if( !l_escaped && l_c == '\\' )
				l_escaped = TRUE;

			else
				{
				l_length++;

				l_escaped = FALSE;
				}
			}

    if( l_c == KLML_EOM )
			{
			ap_parse_ctrl->error = ELML_MESSAGE_TOO_SHORT;
			return FALSE;
			}
		}


	*ap_length = l_length;


    return TRUE;
}


static tlms_bool p_read_ascii( slml_parse_ctrl *ap_parse_ctrl,
								slml_message_element *ap_element,
								char ac_string[],
								int a_max_length )
{
    int l_c;
    int l_length;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( ! (l_c == '\'' || MLML_IS_FIRST_TOKEN_CHAR( l_c ) ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


    a_max_length--;     // Ensure space for terminating '\0' 

	// Note direct call to p_get_char so that contents of string can not be interpreted
	// -as comments

	if( l_c != '\'' )
		{
		for( l_length = 0; MLML_IS_TOKEN_CHAR( l_c ); l_length++ )
			{
			if( l_length >= a_max_length )
				{
				ap_parse_ctrl->error = ELML_PARAMETER_TOO_LONG;
				return FALSE;
				}

			ac_string[l_length] = l_c;
			l_c = p_get_char( ap_parse_ctrl );
			}
		}
	else
		{
		int l_escaped = FALSE;

		l_length = 0;

		while( (l_c = p_get_char( ap_parse_ctrl )) != KLML_EOM && 
						(l_c != '\'' || l_escaped) )
			{
			if( l_length >= a_max_length )
				{
				ap_parse_ctrl->error = ELML_PARAMETER_TOO_LONG;
				return FALSE;
				}

			if( !l_escaped && l_c == '\\' )
				l_escaped = TRUE;

			else if( l_escaped )
				{
				if( l_c == 'r' )
					ac_string[l_length++] = '\r';
				else if( l_c == 'n' )
					ac_string[l_length++] = '\n';
				else if( l_c == 't' )
					ac_string[l_length++] = '\t';
				else    // Also covers case for l_c == '\'' & '\\'
					ac_string[l_length++] = l_c;

				l_escaped = FALSE;
				}

			else
				ac_string[l_length++] = l_c;
			}
		}

    ac_string[l_length] = '\0';


    return TRUE;
}


static tlms_bool p_get_unicode_length( slml_parse_ctrl *ap_parse_ctrl,
										slml_message_element *ap_element,
										int *ap_length )
{
    int l_c;
	int l_escaped = FALSE;
    int l_length = 0;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( l_c != '"' )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


    // Note direct call to p_get_char so that contents of string can not be interpreted
    // -as comments

	while( (l_c = p_get_unicode_char( ap_parse_ctrl )) != KLML_EOM &&
			(l_c != '"' || l_escaped) )
		{
		if( !l_escaped && l_c == '\\' )
			l_escaped = TRUE;

		else
			{
			l_length++;

			l_escaped = FALSE;
			}
		}


    if( l_c == KLML_EOM )
		{
		ap_parse_ctrl->error = ELML_MESSAGE_TOO_SHORT;
		return FALSE;
		}

	*ap_length = l_length;

    return TRUE;
}



static tlms_bool p_read_unicode( slml_parse_ctrl *ap_parse_ctrl,
									slml_message_element *ap_element,
									tlms_unicode ac_string[],
									int a_max_length )
{
    int l_c;
	int l_escaped = FALSE;
    int l_i = 0;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( l_c != '"' )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


    a_max_length--;     // Ensure space for terminating '\0' 


	while( (l_c = p_get_unicode_char( ap_parse_ctrl )) != KLML_EOM && 
						l_i < a_max_length &&
						(l_c != '"' || l_escaped) )
		{
		if( !l_escaped && l_c == '\\' )
			l_escaped = TRUE;

		else if( l_escaped )
			{
			if( l_c == 'r' )
				ac_string[l_i++] = '\r';
			else if( l_c == 'n' )
				ac_string[l_i++] = '\n';
			else if( l_c == 't' )
				ac_string[l_i++] = '\t';
			else    // Also covers case for l_c == '"' & '\\'
				ac_string[l_i++] = l_c;

			l_escaped = FALSE;
			}

		else
			ac_string[l_i++] = l_c;
		}

    ac_string[l_i] = '\0';


    if( l_c == KLML_EOM )
		{
		ap_parse_ctrl->error = ELML_MESSAGE_TOO_SHORT;
		return FALSE;
		}

    else if( l_c != '"' )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_TOO_LONG;
		return FALSE;
		}


    return TRUE;
}


// Base64 table from RFC 1521
//
//           0 A            17 R            34 i            51 z
//           1 B            18 S            35 j            52 0
//           2 C            19 T            36 k            53 1
//           3 D            20 U            37 l            54 2
//           4 E            21 V            38 m            55 3
//           5 F            22 W            39 n            56 4
//           6 G            23 X            40 o            57 5
//           7 H            24 Y            41 p            58 6
//           8 I            25 Z            42 q            59 7
//           9 J            26 a            43 r            60 8
//          10 K            27 b            44 s            61 9
//          11 L            28 c            45 t            62 +
//          12 M            29 d            46 u            63 /
//          13 N            30 e            47 v
//          14 O            31 f            48 w         (pad) =
//          15 P            32 g            49 x
//          16 Q            33 h            50 y

static char p_base64_char_to_6_bits( int a_c )
{
	MLML_ASSERT( isalnum( l_c ) || l_c == '+' || l_c == '/' );

	if( a_c >= 'A' && a_c <= 'Z' )
		return( a_c - 'A' );
	else if( a_c >= 'a' && a_c <= 'z' )
		return( a_c - 'a' + 26 );
	else if( isdigit( a_c ) )
		return( a_c - '0' + 52 );
	else if( a_c == '+' )
		return 62;
	else // if( a_c == '/' )
		return 63;
}


static void p_base64_convert_24bits_to_bytes( char ac_data[], int a_size, int a_accumulated )
{
	if( a_size >= 3 ) ac_data[2] = a_accumulated & 0xff;

	a_accumulated >>= 8;

	if( a_size >= 2 ) ac_data[1] = a_accumulated & 0xff;

	a_accumulated >>= 8;

	if( a_size >= 1 ) ac_data[0] = a_accumulated & 0xff;
}


static tlms_bool p_get_bytes_length( slml_parse_ctrl *ap_parse_ctrl,
											slml_message_element *ap_element,
											int *ap_length )
{
	int l_c;
	int l_number_accumulated;
	int l_at_end = 0;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( l_c != '[' )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	// Must use p_get_char to allow for // in base64.  
	// Allow for multiple lines of input.

	l_number_accumulated = 0;
	while( (l_c = p_get_char( ap_parse_ctrl )) != KLML_EOM && l_c != ']' )
		{
		if( ! l_at_end && ( isalnum( l_c ) || l_c == '+' || l_c == '/' ) )
			l_number_accumulated++;

		else if( l_c == '=' )
			l_at_end = 1;

		else if( ! isspace( l_c ) )
			{
			ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
			return FALSE;
			}
		}

	*ap_length = 3 * (l_number_accumulated / 4);	// 4 chars -> 3 bytes in Base64

	if( l_number_accumulated & 0x03 == 2 )
		*ap_length++;

	else if( l_number_accumulated & 0x03 == 3 )
		*ap_length += 2;


    return TRUE;
}


static tlms_bool p_read_bytes( slml_parse_ctrl *ap_parse_ctrl,
										slml_message_element *ap_element,
										char ac_data[],
										int a_max_length,
										int *ap_result_length )
{
    int l_c;
    int l_i = 0;
	int l_accumulated = 0;
	int l_number_accumulated = 0;
	int l_at_end = 0;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;

	if( l_c != '[' )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	// Must use p_get_char to allow for // in base64
	l_c = p_get_char( ap_parse_ctrl );

	while( l_i<a_max_length && l_c != KLML_EOM && l_c != ']' )
		{
		if( ! l_at_end && ( isalnum( l_c ) || l_c == '+' || l_c == '/' ) )
			{
			l_accumulated <<= 6;
			l_accumulated |= p_base64_char_to_6_bits( l_c );
			if( ++l_number_accumulated == 4 )
				{
				p_base64_convert_24bits_to_bytes( &(ac_data[l_i]), a_max_length - l_i, l_accumulated );

				l_accumulated = l_number_accumulated = 0;
				l_i += 3;
				}
			}

		else if( l_c == '=' )
			++l_at_end;

		else if( ! isspace( l_c ) )
			{
			ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
			return FALSE;
			}

		l_c = p_get_char( ap_parse_ctrl );
		}


	if( l_c == KLML_EOM )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	if( l_number_accumulated == 1 )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	else if( l_number_accumulated >= 2 )	// Can only be 2 or 3
		{
		int l_residual_bytes = 1;

		if( l_number_accumulated == 3 ) l_residual_bytes = 2;

		while( l_number_accumulated++ < 4 )
			l_accumulated <<= 6;

		p_base64_convert_24bits_to_bytes( &(ac_data[l_i]), a_max_length - l_i, l_accumulated );

		l_i += l_residual_bytes;
		}


	*ap_result_length = l_i;

    return TRUE;
}


static tlms_bool p_get_embedded_length( slml_parse_ctrl *ap_parse_ctrl,
								slml_message_element *ap_element,
								int *ap_result_length )
{
    int l_c;
    int l_start = 0;


    if( ! p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c ) )
		return FALSE;


    if( l_c != '(' )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	l_start = p_get_input_location( ap_parse_ctrl );

	p_skip_value( ap_parse_ctrl, l_c );

	*ap_result_length = p_get_input_location( ap_parse_ctrl ) - l_start - 1;
			// Subtract 1 as don't want to count terminating ')'


    return TRUE;
}


static tlms_bool p_read_embedded( slml_parse_ctrl *ap_parse_ctrl,
								slml_message_element *ap_element,
								char ac_data[],
								int a_max_length,
								int *ap_result_length )
{
    int l_c;
    int l_i = 0;
	int l_length;


    if( ! p_get_embedded_length( ap_parse_ctrl, ap_element, &l_length ) )
		return FALSE;


	p_get_first_char( ap_parse_ctrl, ap_element->value_location, &l_c );

	l_c = p_get_char( ap_parse_ctrl );

	for( l_i = 0; l_i<l_length && l_c != KLML_EOM; l_i++ )
		{
		ac_data[l_i] = l_c;
		l_c = p_get_char( ap_parse_ctrl );
		}


    if( l_c == KLML_EOM )
		{
		ap_parse_ctrl->error = ELML_MESSAGE_TOO_SHORT;
		return FALSE;
		}

    else if( l_length > a_max_length )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_TOO_LONG;
		return FALSE;
		}


	*ap_result_length = l_length;


    return TRUE;
}


static void p_free_element_info( slml_message_element *ap_element )
{
    slml_message_element *lp_element_to_free;

    while( ap_element )
        {
        if( ap_element->p_tag )
			free( ap_element->p_tag );

        if( ap_element->p_compound )
            p_free_element_info( ap_element->p_compound );

        lp_element_to_free = ap_element;

        ap_element = ap_element->p_next;

        free( lp_element_to_free );
        }
}


//-----------------------------------------------------------------------------
//                          Start of API functions
//-----------------------------------------------------------------------------


//-----------------------------------------------------------------------------
//              API functions for initiating the parse routines
//-----------------------------------------------------------------------------


// Func name:   flml_is_message_complete
//------------------------------------------------------------------------------
// Parameters:	const char *ap_message - Pointer to the buffer containing the
//							message. 
//				int a_length - The length of the buffer containing the message.
// Returns:		int - The length of the message if complete, or -1 if 
//							incomplete.
// Description: Tests whether the message is complete according to the method
//				of having an unmatched closing brace at the end of the message.

int flml_is_message_complete( const char *ap_message, int a_length )
{
	int l_c;
	slml_parse_ctrl l_parse_ctrl;
	int l_nesting = 0;

	flml_initialise( &l_parse_ctrl );

	l_parse_ctrl.p_message = l_parse_ctrl.p_next = ap_message;
	l_parse_ctrl.p_message_end = ap_message + a_length;


	l_c = p_read( &l_parse_ctrl );


	while( l_c != KLML_EOM )
		{
		if( l_c == '{' )
			l_nesting++;

		else if( l_c == '}' )
			{
			l_nesting--;
			if( l_nesting < 0 )
				return p_get_input_location( &l_parse_ctrl );
			}

		else if( l_c == '=' || isspace( l_c ) )
			{
			// Ignore and move on
			}

		else
			p_skip_value( &l_parse_ctrl, l_c );		// Skip tags as well as values

		
		l_c = p_read( &l_parse_ctrl );
		}


	return -1;
}


// Func name:   flml_initialise
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise
//
// Description: Initialises the user allocated slml_parse_ctrl structure with
//				the details of the message to be parsed.  
//
//				N.B. This function is called as part of flml_parse_message and
//				so it is not normally necessary to call it separately.

tlms_bool flml_initialise( slml_parse_ctrl *ap_parse_ctrl )
{
	ap_parse_ctrl->allocation_failed = FALSE;
	ap_parse_ctrl->root_location.p_element = NULL;
	ap_parse_ctrl->base_location.p_element = NULL;
	ap_parse_ctrl->base_location.p_tag_info = 
			ap_parse_ctrl->root_location.p_tag_info;	// Allow for root to be set already
	ap_parse_ctrl->unget_char = '\0';
	ap_parse_ctrl->recursion = 0;

	ap_parse_ctrl->error = ELML_OK;

	return TRUE;
}


// Func name:   flml_set_tag_info_root
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				slml_tag_info *ap_tag_info - Pointer to the beginning of the
//							structure containing the tagging information.
// Returns:		void
//
// Description: Specifies the base address of the tag info data structure.

void flml_set_tag_info_root( slml_parse_ctrl *ap_parse_ctrl, slml_tag_info *ap_tag_info )
{
	ap_parse_ctrl->root_location.p_tag_info = 
						ap_parse_ctrl->base_location.p_tag_info = ap_tag_info;
}


// Func name:   flml_get_last_error
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
// Returns:		elml_error - The last error code.
//
// Description: Returns the lastest error and resets the error condition to
//				ELML_OK.

elml_error flml_get_last_error( slml_parse_ctrl *ap_parse_ctrl )
{
	elml_error l_last_error = ap_parse_ctrl->error;

	ap_parse_ctrl->error = ELML_OK;

	return l_last_error;
}


// Func name:   flml_parse_message
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				const char *ap_message - Pointer to the encoded message.
//				int a_length - The length of the encoded message.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Parses the message.  This builds a tree representation of the
//				message that the parameter reading functions use to read parts
//				of the message.
//
//				flml_release should be called to release the memory allocated 
//				by this function once all the interaction with the message has
//				been completed.

tlms_bool flml_parse_message( slml_parse_ctrl *ap_parse_ctrl, 
								const char *ap_message, 
								int a_length )
{
	int l_parse_result;

	flml_initialise( ap_parse_ctrl );

	ap_parse_ctrl->p_message = ap_parse_ctrl->p_next = ap_message;
	ap_parse_ctrl->p_message_end = ap_message + a_length;

    ap_parse_ctrl->allocation_failed = FALSE;

	l_parse_result = p_parse( ap_parse_ctrl, &(ap_parse_ctrl->base_location.p_element) );

	ap_parse_ctrl->root_location.p_element = ap_parse_ctrl->base_location.p_element;

	if( l_parse_result == FALSE )
		{
		if( ap_parse_ctrl->allocation_failed && ap_parse_ctrl->error == ELML_OK )
			ap_parse_ctrl->error = ELML_ALLOCATION_FAILED;

		return FALSE;
		}

	else if( ! ap_parse_ctrl->root_location.p_element )
		{
		ap_parse_ctrl->error = ELML_NO_DATA_IN_MESSAGE;
		return FALSE;
		}

    return TRUE;
}


//-----------------------------------------------------------------------------
//        API functions for setting the read position within messages
//-----------------------------------------------------------------------------


// Func name:   flml_set_location
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				slml_location *ap_old_location - Where the previous location 
//							information should be stored.  N.B. This can be
//							NULL if the location information is not required.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Sets the base read location in the tree from which subsequent
//				parameter specifications are made.  If = is the first character
//				then the specification is addressed from the root of the 
//				message tree, otherwise it is addressed from the current
//				location.  By analogy to file directories, it is similar to the 
//				cd command.

tlms_bool flml_set_location( slml_parse_ctrl *ap_parse_ctrl,
									const char *ap_tagging, 
									int a_instance,
									slml_location *ap_old_location )
{
	if( ap_old_location )
		*ap_old_location = ap_parse_ctrl->base_location;

	
	if( *ap_tagging == '=' )
		{
		ap_parse_ctrl->base_location = ap_parse_ctrl->root_location;
		ap_tagging++;
		}


	if( *ap_tagging != '\0' )
		{
		slml_location l_location = ap_parse_ctrl->base_location;

		ap_parse_ctrl->error = p_locate_parameter( &l_location, 
													ap_tagging,
													a_instance );

		if( ap_parse_ctrl->error != ELML_OK )
			return FALSE;

		else if( l_location.p_element == NULL || l_location.p_element->p_compound == NULL )
			{
			ap_parse_ctrl->error = ELML_PARAMETER_NOT_FOUND;
			return FALSE;
			}


		ap_parse_ctrl->base_location.p_element = l_location.p_element->p_compound;
#if LML_USE_COMPILED_TAG_INFO != 0
		ap_parse_ctrl->base_location.p_tag_info = l_location.p_tag_info->p_definition;
#endif
		}


	return TRUE;
}


// Func name:   flml_restore_location
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				slml_location a_base - The base location information being 
//							restored.
// Returns:		slml_location - The base location set at the time the function
//							is called.
//
// Description: Restores the read location to a previous location returned 
//				from any of the other location setting functions.

slml_location flml_restore_location( slml_parse_ctrl *ap_parse_ctrl,
										slml_location a_restore_location )
{
	slml_location l_old_location = ap_parse_ctrl->base_location;

	ap_parse_ctrl->base_location = a_restore_location;

	return l_old_location;
}


//-----------------------------------------------------------------------------
//                   API functions for reading the values
//-----------------------------------------------------------------------------


// Func name:   flml_is_present
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Indicates whether the named parameter is present.

tlms_bool flml_is_present( slml_parse_ctrl *ap_parse_ctrl,
							const char ac_tagging[],
							int a_instance )
{
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( lp_element )
        return TRUE;

    return FALSE;
}


// Func name:   flml_read_void
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the void value associated with the named parameter.  In 
//				practice, this just checks for the presence of the void's tag.

tlms_bool flml_read_void( slml_parse_ctrl *ap_parse_ctrl,
						const char ac_tagging[],
						int a_instance )
{
	if( ! flml_is_present( ap_parse_ctrl, ac_tagging, a_instance ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_NOT_FOUND;
		return FALSE;
		}

	return TRUE;
}


// Func name:   flml_read_boolean
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int *ap_boolean - Pointer to where the result is to be placed,
//							which will either be TRUE or FALSE.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the boolean value associated with the named parameter.

tlms_bool flml_read_boolean( slml_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								int *ap_boolean )
{
    int l_c;
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;

	if( ! p_get_first_char( ap_parse_ctrl, lp_element->value_location, &l_c ) )
		return FALSE;

    if( l_c == 'T' || l_c == 't' )
        *ap_boolean = TRUE;

    else if( l_c == 'F' || l_c == 'f' )
        *ap_boolean = FALSE;

    else
		{
        ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

    return TRUE;
}


// Func name:   flml_read_integer
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int *ap_integer - Pointer to where the integer result is to be 
//							placed.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the integer value associated with the named parameter.

tlms_bool flml_read_integer( slml_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								int *ap_integer )
{
    int l_c;
    int l_value = 0;
    int l_positive = TRUE;
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;


	if( ! p_get_first_char( ap_parse_ctrl, lp_element->value_location, &l_c ) )
		return FALSE;

    if( l_c == '-' )
        {
        l_positive = FALSE;

        l_c = p_read( ap_parse_ctrl );
        }

    if( ! isdigit( l_c ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

    while( isdigit( l_c ) )
        {
        l_value *= 10;

        l_value += l_c - '0';

        l_c = p_read( ap_parse_ctrl );
        }

    if( l_positive )
        *ap_integer = l_value;
    else
        *ap_integer = -l_value;


	if( ! MLMS_AFTER_VALUE_CHAR( l_c ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


    return TRUE;
}


// Func name:   flml_read_float
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				float *ap_float - Pointer to where the float result is to be 
//							placed.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the float value associated with the named parameter.

tlms_bool flml_read_float( slml_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								tlms_float *ap_float )
{
	double l_double;

	tlms_bool l_return = flml_read_double( ap_parse_ctrl, ac_tagging,
											a_instance,&l_double );

	*ap_float = (tlms_float)l_double;

	return l_return;
}


// Func name:   flml_read_double
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int *ap_double - Pointer to where the integer result is to be 
//							placed.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the double value associated with the named parameter.

tlms_bool flml_read_double( slml_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								tlms_double *ap_double )
{
	int l_c;
	int l_i;
	char lc_value[50];
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( ! lp_element )
		return FALSE;


	if( ! p_get_first_char( ap_parse_ctrl, lp_element->value_location, &l_c ) )
		return FALSE;


	for( l_i=0; l_i < sizeof( lc_value ) - 1 && MLML_IS_TOKEN_CHAR( l_c ); l_i++ )
		{
		lc_value[l_i] = l_c;
		l_c = p_read( ap_parse_ctrl );
		}

	lc_value[l_i] = '\0';

	if( l_i >= sizeof( lc_value ) - 1 )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	if( strcmp( lc_value, "NaN" ) == 0 )		// See portability note: IEEE754
		*ap_double = 0.0 / p_float_zero();
	else if( strcmp( lc_value, "INF" ) == 0 ) 
		*ap_double = 1.0 / p_float_zero();
	else if( strcmp( lc_value, "-INF" ) == 0 ) 
		*ap_double = -1.0 / p_float_zero();
	else if( p_is_valid_float( lc_value ) ) 
		*ap_double = atof( lc_value );
	else
		{
        ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	return TRUE;
}


// Func name:   flml_read_ipv4
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				unsigned char ac_addr[4] - Where the 4 bytes of the IPv4 
//							network address are to be placed.  This is in
//							network byte order.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the IPv4 address value associated with the named 
//				parameter.

tlms_bool flml_read_ipv4( slml_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								unsigned char ac_addr[4] )
{
	int l_c;
	unsigned int l_int1, l_int2, l_int3, l_int4;
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( ! lp_element )
		return FALSE;


    p_set_input_location( ap_parse_ctrl, lp_element->value_location );

	if( ! p_read_pos_integer( ap_parse_ctrl, &l_int1 ) ||
					'.' != p_read( ap_parse_ctrl ) ||
				! p_read_pos_integer( ap_parse_ctrl, &l_int2 ) ||
					'.' != p_read( ap_parse_ctrl ) ||
				! p_read_pos_integer( ap_parse_ctrl, &l_int3 ) ||
					'.' != p_read( ap_parse_ctrl ) ||
				! p_read_pos_integer( ap_parse_ctrl, &l_int4 ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	if( l_int1 > 255 || l_int2 > 255 || l_int3 > 255 || l_int4 > 255 )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	l_c = p_read( ap_parse_ctrl );

	if( ! MLMS_AFTER_VALUE_CHAR( l_c ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	ac_addr[0] = l_int1;
	ac_addr[1] = l_int2;
	ac_addr[2] = l_int3;
	ac_addr[3] = l_int4;


    return TRUE;
}


// Func name:   flml_read_date
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				tlms_date *ap_date - Pointer to where the date is to be placed.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the date value associated with the named parameter.

tlms_bool flml_read_date( slml_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								tlms_date *ap_date )
{
	int l_c;
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( ! lp_element )
		return FALSE;

    p_set_input_location( ap_parse_ctrl, lp_element->value_location );

	if( ! p_read_pos_integer( ap_parse_ctrl, &(ap_date->year) ) ||
					'-' != p_read( ap_parse_ctrl ) ||
				! p_read_pos_integer( ap_parse_ctrl, &(ap_date->month) ) ||
					'-' != p_read( ap_parse_ctrl ) ||
				! p_read_pos_integer( ap_parse_ctrl, &(ap_date->day) ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	if( ap_date->month < 1 || ap_date->month > 12 ||
				ap_date->day < 1 || ap_date->day > 31 )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	l_c = p_read( ap_parse_ctrl );

	if( ! MLMS_AFTER_VALUE_CHAR( l_c ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


    return TRUE;
}


// Func name:   flml_read_time
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				tlms_date *ap_time - Pointer to where the time is to be placed.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the time value associated with the named parameter.

tlms_bool flml_read_time( slml_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								tlms_time *ap_time )
{
	int l_c;
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( ! lp_element )
		return FALSE;


    p_set_input_location( ap_parse_ctrl, lp_element->value_location );

	if( ! p_read_pos_integer( ap_parse_ctrl, &(ap_time->hour) ) ||
					':' != p_read( ap_parse_ctrl ) ||
				! p_read_pos_integer( ap_parse_ctrl, &(ap_time->min) ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	if( ap_time->hour < 0 || ap_time->hour > 23 ||
				ap_time->min < 0 || ap_time->min > 59 )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}
					
	ap_time->sec = 0;

	l_c = p_read( ap_parse_ctrl );

	if( isspace( l_c ) )
		return TRUE;

	if( ':' != l_c ||
				! p_read_pos_integer( ap_parse_ctrl, &(ap_time->sec) ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	if( ap_time->sec < 0 || ap_time->sec > 59 )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}

	l_c = p_read( ap_parse_ctrl );

	if( ! MLMS_AFTER_VALUE_CHAR( l_c ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


    return TRUE;
}


// Func name:   flml_read_oid
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				slml_oid *ap_oid - Where the OID is to be placed.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the OID value associated with the named parameter.

tlms_bool flml_read_oid( slml_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								tlms_oid *ap_oid )
{
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( ! lp_element )
		return FALSE;


    p_set_input_location( ap_parse_ctrl, lp_element->value_location );

	ap_oid->size = 0;

	if( ! p_read_pos_integer( ap_parse_ctrl, &(ap_oid->value[ap_oid->size++]) ) )
		{
		ap_parse_ctrl->error = ELML_PARAMETER_WRONG_TYPE;
		return FALSE;
		}


	while( '~' == p_read( ap_parse_ctrl ) &&
			p_read_pos_integer( ap_parse_ctrl, &(ap_oid->value[ap_oid->size++]) ) )
		{
		if( ap_oid->size >= sizeof( ap_oid->value ) / sizeof( ap_oid->value[0] ) )
			{
			ap_parse_ctrl->error = ELML_PARAMETER_TOO_LONG;
			return FALSE;
			}
		}

    return TRUE;
}


// Func name:   flml_read_ascii
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char ac_string[] - Where the ASCII string is to be placed.
//				int a_max_length - The length of the ac_string array.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the unquoted or quoted ascii string value associated with 
//				the named parameter.

tlms_bool flml_read_ascii( slml_parse_ctrl *ap_parse_ctrl,
							const char ac_tagging[],
							int a_instance, 
							char ac_string[],
							int a_max_length )
{
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;

	return p_read_ascii( ap_parse_ctrl, lp_element, ac_string, a_max_length );
}


// Func name:   flml_read_ascii_dynamic
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char **app_string - Pointer to where the pointer to the 
//							dynamically allocated string buffer is to be 
//							placed.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the unquoted or quoted ascii string value associated with 
//				the named parameter into an array that is dynamically allocated
//				by the library.  
//
//				It is the responsibility of the calling code to free this memory
//				when it is no longer required.

tlms_bool flml_read_ascii_dynamic( slml_parse_ctrl *ap_parse_ctrl,
										const char ac_tagging[],
										int a_instance, 
										char **app_string )
{
	int l_length;
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;


	*app_string = NULL;


	if( ! p_get_ascii_length( ap_parse_ctrl, lp_element, &l_length ) )
		return FALSE;

	
	l_length++;	// Allow for terminating \0

	*app_string = (char *)flml_alloc( ap_parse_ctrl, l_length * sizeof(char) );

	if( *app_string == NULL )
		{
		ap_parse_ctrl->error = ELML_ALLOCATION_FAILED;
		return FALSE;
		}


	return p_read_ascii( ap_parse_ctrl, lp_element, *app_string, l_length );
}


// Func name:   flml_read_unicode
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				tlms_unicode ac_string[] - Where the Unicode string is to be 
//							placed.
//				int a_max_length - The length of the ac_string array.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the unicode string value associated with the named 
//				parameter.

tlms_bool flml_read_unicode( slml_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								tlms_unicode ac_string[],
								int a_max_length )
{
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;


	return p_read_unicode( ap_parse_ctrl, lp_element, ac_string, a_max_length );
}


// Func name:   flml_read_unicode_dynamic
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				tlms_unicode **app_string - Pointer to where the pointer to the 
//							dynamically allocated string buffer is to be 
//							placed.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the unicode string value associated with the named 
//				parameter into an array that is dynamically allocated by the 
//				library.  
//
//				It is the responsibility of the calling code to free this memory
//				when it is no longer required.

tlms_bool flml_read_unicode_dynamic( slml_parse_ctrl *ap_parse_ctrl,
										const char ac_tagging[],
										int a_instance, 
										tlms_unicode **app_string )
{
	int l_length;
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;


	*app_string = NULL;


	if( ! p_get_unicode_length( ap_parse_ctrl, lp_element, &l_length ) )
		return FALSE;
	
	l_length++;	// Allow for terminating \0

	*app_string = (tlms_unicode *)flml_alloc( ap_parse_ctrl, l_length * sizeof(tlms_unicode) );

	if( *app_string == NULL )
		{
		ap_parse_ctrl->error = ELML_ALLOCATION_FAILED;
		return FALSE;
		}


	return p_read_unicode( ap_parse_ctrl, lp_element, *app_string, l_length );
}


// Func name:   flml_read_const
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char ac_string[] - Where the const ASCII string is to be placed.
//				int a_max_length - The length of the ac_string array.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the const value associated with the named parameter.

tlms_bool flml_read_const( slml_parse_ctrl *ap_parse_ctrl,
							const char ac_tagging[],
							int a_instance, 
							char ac_string[],
							int a_max_length )
{
	return flml_read_ascii( ap_parse_ctrl,
								ac_tagging,
								a_instance, 
								ac_string,
								a_max_length );
}


// Func name:   flml_bytes
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char ac_data[] - Where the bytes data is to be placed.
//				int a_max_length - The length of the ac_data array.
//				int *ap_result_length - Where the resulting number of bytes 
//							stored in ac_data should be stored.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the bytes value associated with the named parameter.

tlms_bool flml_read_bytes( slml_parse_ctrl *ap_parse_ctrl,
							const char ac_tagging[],
							int a_instance, 
							char ac_data[],
							int a_max_length,
							int *ap_result_length )
{
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;


	return p_read_bytes( ap_parse_ctrl, lp_element, ac_data, a_max_length, ap_result_length );
}


// Func name:   flml_read_bytes_dynamic
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char **app_data - Pointer to where the pointer to the 
//							dynamically allocated bytes data buffer is to be 
//							placed.
//				int *ap_result_length - Where the resulting number of bytes 
//							stored in app_data should be stored.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the bytes value associated with the named parameter 
//				into an array that is dynamically allocated by the library.
//
//				It is the responsibility of the calling code to free this memory
//				when it is no longer required.

tlms_bool flml_read_bytes_dynamic( slml_parse_ctrl *ap_parse_ctrl,
										const char ac_tagging[],
										int a_instance, 
										char **app_data,
										int *ap_result_length )
{
	int l_length;
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( ! lp_element )
		return FALSE;


	*app_data = NULL;


	if( ! p_get_bytes_length( ap_parse_ctrl, lp_element, &l_length ) )
		return FALSE;

	
	*app_data = (char *)flml_alloc( ap_parse_ctrl, l_length * sizeof(char) );

	if( *app_data == NULL )
		{
		ap_parse_ctrl->error = ELML_ALLOCATION_FAILED;
		return FALSE;
		}


	return p_read_bytes( ap_parse_ctrl, lp_element, *app_data, l_length, ap_result_length );
}


// Func name:   flml_read_embedded
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char ac_data[] - Where the embedded data is to be placed.
//				int a_max_length - The length of the ac_data array.
//				int *ap_result_length - Where the resulting number of bytes 
//							stored in ac_data should be stored.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the embedded value associated with the named parameter.

tlms_bool flml_read_embedded( slml_parse_ctrl *ap_parse_ctrl,
								const char ac_tagging[],
								int a_instance, 
								char ac_data[],
								int a_max_length,
								int *ap_result_length )
{
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );


    if( ! lp_element )
		return FALSE;


	return p_read_embedded( ap_parse_ctrl, lp_element, ac_data, a_max_length, ap_result_length );
}


// Func name:   flml_read_embedded_dynamic
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char **app_data - Pointer to where the pointer to the 
//							dynamically allocated data buffer is to be 
//							placed.
//				int *ap_result_length - Where the resulting number of bytes 
//							stored in app_data should be stored.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the embedded value associated with the named 
//				parameter into an array that is dynamically allocated by the 
//				library.  
//
//				It is the responsibility of the calling code to free this memory
//				when it is no longer required.

tlms_bool flml_read_embedded_dynamic( slml_parse_ctrl *ap_parse_ctrl,
										const char ac_tagging[],
										int a_instance, 
										char **app_data,
										int *ap_result_length )
{
	int l_length;
    slml_message_element *lp_element = p_locate( ap_parse_ctrl,
													ac_tagging,
													a_instance );

    if( lp_element == NULL )
		return FALSE;


	*app_data = NULL;


	if( ! p_get_embedded_length( ap_parse_ctrl, lp_element, &l_length ) )
		return FALSE;

	
	*app_data = (char *)flml_alloc( ap_parse_ctrl, l_length * sizeof(char) );

	if( *app_data == NULL )
		{
		ap_parse_ctrl->error = ELML_ALLOCATION_FAILED;
		return FALSE;
		}


	return p_read_embedded( ap_parse_ctrl, lp_element, *app_data, l_length, ap_result_length );
}


// Func name:   flml_read_union
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				char ac_string[] - Where the string specified in the union is 
//							to be placed.
//				int a_max_length - The length of the ac_string array.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Reads the selected option of the union associated with the 
//				named parameter.

tlms_bool flml_read_union( slml_parse_ctrl *ap_parse_ctrl,
									const char ac_tagging[],
									int a_instance, 
									char ac_string[],
									int a_max_length )
{
	return flml_read_ascii( ap_parse_ctrl,
								ac_tagging,
								a_instance, 
								ac_string,
								a_max_length );
}


// Func name:   flml_read_enumerated_union
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
//				char ac_tagging[] - Tagging information for the parameter that 
//							is to be read.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				int a_instance - The instance of the tag that is being looked 
//							for.  See the section 
//							<Tagging Information> for more information on how 
//							to specify tags.
//				slml_tag_translation ac_tag_translation[] - Pointer to the 
//							array mapping between ascii words and integer 
//							values to be used for the enumeration.
// Returns:		int - The integer corresponding to the union value 
//							present.  If there is not an entry in the 
//							translation table, then -1 is returned.
//
// Description: Reads the selected option of the union associated with the 
//				named parameter and converts it to an integer value according to
//				the contents of the array ac_tag_translation.  This is indended
//				for use in switch statements.

int flml_read_enumerated_union( slml_parse_ctrl *ap_parse_ctrl,
									const char ac_tagging[],
									int a_instance, 
									slml_tag_translation ac_tag_translation[] )
{
	char lc_union_option[KLML_MAX_TAG];

	if( flml_read_union( ap_parse_ctrl,  
							ac_tagging, a_instance, 
							lc_union_option, sizeof(lc_union_option) ) )
		{
		int l_index;
		char *lp_sought = lc_union_option;

		if( isdigit( lc_union_option[0] ) ) 
			lp_sought = "";

		for( l_index=0; ac_tag_translation[l_index].text_tag != NULL; l_index++ )
			{
			if( ac_tag_translation[l_index].text_tag[0] == lp_sought[0] &&
					( lp_sought[0] == '\0' ||
						strcmp( &(ac_tag_translation[l_index].text_tag[1]), 
									&(lp_sought[1]) ) == 0 ) )
				{
				return( ac_tag_translation[l_index].enumerated_value );
				}
			}

		}

	return -1;
}


//-----------------------------------------------------------------------------
//                         Functions for clearing up
//-----------------------------------------------------------------------------


// Func name:   flml_release
//------------------------------------------------------------------------------
// Parameters:	slml_parse_ctrl *ap_parse_ctrl - Pointer to the parse control 
//							block.
// Returns:		void
//
// Description: Releases all the resources associated with parsing the message.
//				This function must be called when the code have finished with
//				a parsed message to avoid memory leaks. 

void flml_release( slml_parse_ctrl *ap_parse_ctrl )
{
	p_free_element_info( ap_parse_ctrl->root_location.p_element );

	ap_parse_ctrl->root_location.p_element = 
					ap_parse_ctrl->base_location.p_element = NULL;
}


//-----------------------------------------------------------------------------
//                      Functions to support encoding
//-----------------------------------------------------------------------------

//   UCS-2 range (hex.)		UTF-8 octet sequence (binary)
//   0000-007F				0xxxxxxx
//   0080-07FF				110xxxxx 10xxxxxx
//   0800-FFFF				1110xxxx 10xxxxxx 10xxxxxx

static int p_unicode_to_escaped_utf8( char ac_utf8[3], tlms_unicode a_unicode_char )
{
	if( a_unicode_char == '\\' || a_unicode_char == '"' )
		{
		ac_utf8[0] = '\\';
		ac_utf8[1] = (char)a_unicode_char;
		return 2;
		}

	else if( a_unicode_char <= 0x7f )
		{
		ac_utf8[0] = (char)a_unicode_char;
		return 1;
		}

	else if( a_unicode_char <= 0x7ff )
		{
		// 2 bytes
		ac_utf8[1] = ((char)a_unicode_char & 0x3f) | 0x80;
		a_unicode_char >>= 6;
		ac_utf8[0] = ((char)a_unicode_char & 0x1f) | 0xC0;
		return 2;
		}

	else
		{
		// 3 bytes
		ac_utf8[2] = ((char)a_unicode_char & 0x3f) | 0x80;
		a_unicode_char >>= 6;
		ac_utf8[1] = ((char)a_unicode_char & 0x3f) | 0x80;
		a_unicode_char >>= 6;
		ac_utf8[0] = ((char)a_unicode_char & 0x0f) | 0xE0;
		return 3;
		}
}


static char p_base64_6_bits_to_base64_char( int a_c )
{
	MLML_ASSERT( a_c <= 0x00ffffff );

	if( a_c >= 0 && a_c <= 25 )
		return( a_c + 'A' );
	else if( a_c >= 26 && a_c <= 51 )
		return( a_c - 26 + 'a' );
	else if( a_c >= 52 && a_c <= 61 )
		return( a_c - 52 + '0' );
	else if( a_c == 62 )
		return '+';
	else // if( a_c == 63 )
		return '/';
}


static void p_base64_convert_24bits_to_4x6bits( char ac_data[], int a_accumulated )
{
	ac_data[3] = p_base64_6_bits_to_base64_char( a_accumulated & 0x3f );

	a_accumulated >>= 6;

	ac_data[2] = p_base64_6_bits_to_base64_char( a_accumulated & 0x3f );

	a_accumulated >>= 6;

	ac_data[1] = p_base64_6_bits_to_base64_char( a_accumulated & 0x3f );

	a_accumulated >>= 6;

	ac_data[0] = p_base64_6_bits_to_base64_char( a_accumulated & 0x3f );
}


static tlms_bool p_is_well_formed( slml_parse_ctrl *ap_parse_ctrl, int a_nest_char )
{
	int l_c;
	int l_past_first_item = FALSE;
	int l_last_is_value = FALSE;

	if( ap_parse_ctrl->recursion++ > KLML_MAX_RECURSIONS )
		{
		ap_parse_ctrl->error = ELML_TOO_MANY_RECURSIONS;
		return FALSE;
		}


	for(;;)
		{
		l_c = p_read( ap_parse_ctrl );
		while( isspace( l_c ) )
			l_c = p_read( ap_parse_ctrl );

		if( l_c == '(' || l_c == '{' )
			{
			if( ! p_is_well_formed( ap_parse_ctrl, l_c ) )
				return FALSE;
			}
		
		else if( l_c == ')' )
			{
			if( a_nest_char == '(' )
				break;
			else
				return FALSE;
			}

		else if( l_c == '}' )
			{
			if( ap_parse_ctrl->recursion == 1 && a_nest_char == '\0' )
				{
				// Assume '}' marks end of message.  Check for stuff afterwards.
				while( (l_c = p_read( ap_parse_ctrl )) != KLML_EOM )
					{
					if( !isspace( l_c ) )
						return FALSE;
					}

				break;
				}

			else if( a_nest_char == '{' )
				break;
			else
				return FALSE;
			}

		else if( l_c == '=' )
			{
			if( l_last_is_value )
				return FALSE;
			}

		else if( MLML_IS_FIRST_TAG_CHAR( l_c ) || MLML_IS_VALUE_START( l_c ) )
			{
			if( MLML_IS_VALUE_START( l_c ) )
				l_last_is_value = TRUE;
			else
				l_last_is_value = FALSE;

			p_skip_value( ap_parse_ctrl, l_c );

			l_past_first_item = TRUE;
			}

		else if( l_c == ',' )
			{
			if( ! l_past_first_item )
				return FALSE;
			}

		else if( l_c == KLML_EOM )
			{
			if( ap_parse_ctrl->recursion == 1 )
				return TRUE;

			return FALSE;
			}

		else
			return FALSE;
		}


	ap_parse_ctrl->recursion--;

	return TRUE;
}


// Func name:   flml_encode_float
//------------------------------------------------------------------------------
// Parameters:	char ac_encoded[] - Where the text representation of the double
//							value is to be placed.
//				tlms_double a_value - The double value to be encoded.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Converts a double value into an ascii string taking into 
//				account possible conversion to NaN, INF and -INF.

tlms_bool flml_encode_float( char ac_encoded_float[], 
								tlms_double a_value )
{
	int l_float_class = _fpclass( a_value );	// See portability note: IEEE754

	if( l_float_class & _FPCLASS_SNAN || l_float_class & _FPCLASS_QNAN )
		strcpy( ac_encoded_float, "NaN" );
	else if( l_float_class & _FPCLASS_NINF )
		strcpy( ac_encoded_float, "-INF" );
	else if( l_float_class & _FPCLASS_NZ )
		strcpy( ac_encoded_float, "-0" );
	else if( l_float_class & _FPCLASS_PZ )
		strcpy( ac_encoded_float, "0" );
	else if( l_float_class & _FPCLASS_PINF )
		strcpy( ac_encoded_float, "INF" );
	else
		sprintf( ac_encoded_float, "%g", a_value );

	return TRUE;
}


// Func name:   flml_encode_get_ascii_length
//------------------------------------------------------------------------------
// Parameters:	const char ac_ascii[] - Array on which to calculate the 
//							encoded length.
// Returns:		int - The length of the item when encoded in Lumas format plus
//				space for a terminating \0.
//
// Description: Returns the number of bytes needed to store an encoded ascii
//				string.  This includes the leading and trailing ' characters,
//				space for any escaping that needs to be done, and a terminating 
//				\0.

int flml_encode_get_ascii_length( const char ac_ascii[] )
{
	int l_ascii_length = 0;
	int l_i;

	for( l_i=0; ac_ascii[l_i] != '\0'; l_i++ )
		if( ac_ascii[l_i] == '\\' || ac_ascii[l_i] == '\'' )
			l_ascii_length += 2;
		else
			l_ascii_length++;

	l_ascii_length += 1 + 2;		// +1 for terminating \0, +2 for ""

	return l_ascii_length;
}


// Func name:   flml_encode_ascii
//------------------------------------------------------------------------------
// Parameters:	char ac_encoded_ascii[] - Where the encoded ASCII string is to
//							be placed.
//				int a_encoded_ascii_max_length - The maximum size of the 
//							ac_encoded_ascii buffer.
//				const char ac_ascii[] - The ASCII string to be encoded.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Converts an ascii string into an encoded ascii
//				string.  This includes the leading and trailing ' characters,
//				space for any escaping that needs to be done, and a terminating 
//				\0.

tlms_bool flml_encode_ascii( char ac_encoded_ascii[], 
								int a_encoded_ascii_max_length,
								const char ac_ascii[] )
{
	int l_i;
	int l_escaped_ascii_length = 0;


	ac_encoded_ascii[l_escaped_ascii_length++] = '\'';

	for( l_i=0; ac_ascii[l_i] != '\0' && 
					l_escaped_ascii_length < a_encoded_ascii_max_length; 
				l_i++ )
		{
		if( ac_ascii[l_i] == '\\' || ac_ascii[l_i] == '\'' )
			ac_encoded_ascii[l_escaped_ascii_length++] = '\\';

		ac_encoded_ascii[l_escaped_ascii_length++] = ac_ascii[l_i];
		}


	if( l_escaped_ascii_length + 2 > a_encoded_ascii_max_length )
		{
		return FALSE;
		}

	ac_encoded_ascii[l_escaped_ascii_length++] = '\'';
	ac_encoded_ascii[l_escaped_ascii_length++] = '\0';

	return TRUE;
}


// Func name:   flml_encode_get_unicode_length
//------------------------------------------------------------------------------
// Parameters:	const tlms_unicode ac_unicode[] - Array on which to calculate 
//							the encoded length.
// Returns:		int - The length of the item when encoded in Lumas format plus
//				space for a terminating \0.
//
// Description: Returns the number of bytes needed to store an encoded Unicode
//				string.  This includes the leading and trailing " characters,
//				space for any UTF-8 encoding that needs to be done, and a 
//				terminating \0.

int flml_encode_get_unicode_length( const tlms_unicode ac_unicode[] )
{
	int l_utf8_length = 0;
	char lc_unused[3];
	int l_i;

	for( l_i=0; ac_unicode[l_i] != 0; l_i++ )
		l_utf8_length += p_unicode_to_escaped_utf8( lc_unused, ac_unicode[l_i] );

	l_utf8_length += 1 + 2;		// +1 for terminating \0, +2 for ""

	return l_utf8_length;
}


// Func name:   flml_encode_unicode
//------------------------------------------------------------------------------
// Parameters:	char ac_encoded_unicode[] - Where the encoded byte Unicode 
//							string is to be placed.
//				int a_encoded_unicode_max_length - The maximum size of the 
//							ac_encoded_unicode buffer.
//				const tlms_unicode ac_unicode[] - The Unicode string to be 
//							encoded.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Converts a Unicode string into the encoded UTF-8 format for
//				outputing on the wire. This includes the leading and trailing 
//				" characters, space for any escaping that needs to be done, 
//				and a terminating \0.

tlms_bool flml_encode_unicode( char ac_encoded_unicode[], 
								int a_encoded_unicode_max_length,
								const tlms_unicode ac_unicode[] )
{
	int l_total_utf8_length = 0;
	char lc_converted_char[3];
	int l_utf8_length;
	int l_i;
	int l_j;

	ac_encoded_unicode[l_total_utf8_length++] = '"';

	for( l_i=0; ac_unicode[l_i] != 0; l_i++ )
		{
		l_utf8_length = p_unicode_to_escaped_utf8( lc_converted_char, ac_unicode[l_i] );
		if( l_total_utf8_length + l_utf8_length <= a_encoded_unicode_max_length )
			{
			for( l_j=0; l_j<l_utf8_length; l_j++ )
				ac_encoded_unicode[l_total_utf8_length++] = lc_converted_char[l_j]; 
			}
		else
			{
			return FALSE;
			}
		}

	if( l_total_utf8_length + 2 > a_encoded_unicode_max_length )
		{
		return FALSE;
		}
		
	ac_encoded_unicode[l_total_utf8_length++] = '"';
	ac_encoded_unicode[l_total_utf8_length++] = '\0';

	return TRUE;
}


#define KLML_BYTES_NL 40	// Start bytes encoding on new line if longer than this

// Func name:   flml_encode_get_bytes_length
//------------------------------------------------------------------------------
// Parameters:	int a_data_length - The length of the unencoded binary data.
// Returns:		int - The length of the item when encoded in Lumas format plus
//				space for a terminating \0.
//
// Description: Returns the number of bytes needed to store a base64 encoded
//				array of a_data_length bytes.  This includes space for the
//				leading and terminating [ ] characters, space for line 
//				breaks every 76 output characters, plus space for a terminating
//				\0 character.

int flml_encode_get_bytes_length( int a_data_length )
{
	int l_data_length = 4 * ( (a_data_length + 2) / 3 );	// The number of chars to encode the data
	l_data_length += l_data_length / 76;					// Line break every 76 chars
	if( l_data_length > KLML_BYTES_NL ) ++l_data_length;	// Start on new line if long
	return l_data_length + 1 + 2;							// +1 for \0 terminator, +2 for []
}


// Func name:   flml_encode_bytes
//------------------------------------------------------------------------------
// Parameters:	char ac_base64[] - Where the encoded byte string is to
//							be placed.
//				int a_max_length - The size of the ac_base64 buffer in bytes.
//				const char *ap_data - Pointer to the data to be encoded. 
//				int a_length - The length of the data in ap_data to be encoded.
// Returns:		tlms_bool - TRUE if successful, FALSE otherwise.  Information
//							on why the function failed can be obtained by 
//							calling <flml_get_last_error>.
//
// Description: Convert binary data stored in a bytes item into base64 encoded
//				data suitable for outputing to an encoded message. The result 
//				includes the leading and terminating [ ] characters, line 
//				breaks every 76 output characters, plus a terminating
//				\0 character.
//

tlms_bool flml_encode_bytes( char ac_base64[],
								int a_max_length,
								const char *ap_data,
								int a_length )
{
    int l_in = 0;
	int l_out = 0;
	int l_accumulated = 0;
	int l_number_accumulated = 0;
	int l_line_length = 0;

	int l_output_length = flml_encode_get_bytes_length( a_length );

	
	if( l_output_length > a_max_length )
		{
		return FALSE;
		}

	ac_base64[l_out++] = '[';

	if( l_output_length > KLML_BYTES_NL )
		ac_base64[l_out++] = '\n';

	for( l_in=0; l_in < a_length; l_in++ )
		{
		l_accumulated <<= 8;
		l_accumulated |= *ap_data++;
		if( ++l_number_accumulated == 3 )
			{
			p_base64_convert_24bits_to_4x6bits( &(ac_base64[l_out]), l_accumulated );

			l_accumulated = l_number_accumulated = 0;
			l_out += 4;

			if( ++l_line_length >= 76 / 4 )
				{
				ac_base64[l_out++] = '\n';
				l_line_length = 0;
				}
			}
		}


	if( l_number_accumulated == 1 )
		{
		l_accumulated <<= 2 * 8;

		p_base64_convert_24bits_to_4x6bits( &(ac_base64[l_out]), l_accumulated );

		ac_base64[l_out+2] = ac_base64[l_out+3] = '=';

		l_out += 4;
		}

	else if( l_number_accumulated == 2 )
		{
		l_accumulated <<= 8;

		p_base64_convert_24bits_to_4x6bits( &(ac_base64[l_out]), l_accumulated );

		ac_base64[l_out+3] = '=';

		l_out += 4;
		}

	ac_base64[l_out++] = ']';
	ac_base64[l_out++] = '\0';

    return TRUE;
}


// Func name:   flml_is_well_formed
//------------------------------------------------------------------------------
// Parameters:	const char *ap_message - Pointer to the buffer containing the
//							message. 
//				int a_length - The length of the message in the buffer.
// Returns:		tlms_bool - TRUE if the message is well formed, FALSE otherwise.
// Description: Tests whether the message is well formed from a syntactic point
//				of view.  (Note that even if the message passes this test, it
//				may still be semantically ill-formed, and may not conform to 
//				the intended message definition.)
//
//				This function is useful to test that locally generated messages 
//				have a valid Lumas format before putting them on the wire.  It 
//				can also be used to test whether received messages have a 
//				valid format.

tlms_bool flml_is_well_formed(  const char *ap_message, int a_length )
{
	slml_parse_ctrl l_parse_ctrl;

	flml_initialise( &l_parse_ctrl );

	l_parse_ctrl.p_message = l_parse_ctrl.p_next = ap_message;
	l_parse_ctrl.p_message_end = ap_message + a_length;

	return p_is_well_formed( &l_parse_ctrl, '\0' );
}



//------------------------------------------------------------------------------
//             Copyright (c) 2002-2003, Tech-Know-Ware Ltd.
//------------------------------------------------------------------------------
